/*
 * Software License Agreement (BSD License)
 *
 * Copyright (c) 2010, Kyoto University and Honda Motor Co.,Ltd. All rights reserved.
 *
 * HARK was developed by researchers in Okuno Laboratory at the Kyoto University and 
 * Honda Research Institute Japan Co.,Ltd.
 *
 * Redistribution and use in source and binary forms, with or without modification, are 
 * permitted provided that the following conditions are met:
 *
 *  * Redistributions of source code must retain the above copyright notice, this list 
 *    of conditions and the following disclaimer.
 *  * Redistributions in binary form must reproduce the above copyright notice, this 
 *    list of conditions and the following disclaimer in the documentation and/or other 
 *    materials provided with the distribution.
 *  * Neither the name of Kyoto University and Honda Motor Co.,Ltd., Inc. nor the names 
 *    of its contributors may be used to endorse or promote products derived from this 
 *    software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY 
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES 
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT 
 * SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCI-
 * DENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, 
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSI-
 * NESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTR-
 * ACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY 
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "BufferedNode.h"
#include "operators.h"
#include "CvfdImage.h"
#include "Map.h"

using namespace std;
using namespace FD;

class CvfdhMorphologyEx;
   
DECLARE_NODE(CvfdhMorphologyEx)
/*Node
 *
 * @name CvfdhMorphologyEx
 * @category FDOPENCV_HARK:Transform
 * @description CvfdhMorphologyEx
 *
 * @input_name IMAGEIN
 * @input_type any
 * @input_description Input image. The type should be CvfdImage or Map<int, ObjectRef>.
 *
 * @output_name IMAGEOUT
 * @output_type any
 * @output_description Output image. The type should be CvfdImage or Map<int, ObjectRef>.
 *
 * @parameter_name RECT_SIZE_X
 * @parameter_type int
 * @parameter_value 9
 * @parameter_description Rectangle window size in x direction
 *
 * @parameter_name RECT_SIZE_Y
 * @parameter_type int
 * @parameter_value 9
 * @parameter_description Rectangle window size in y direction
 *
 * @parameter_name ANCHOR_POINT_X
 * @parameter_type int
 * @parameter_value 4
 * @parameter_description Anchor point of the rectangle window in x direction
 *
 * @parameter_name ANCHOR_POINT_Y
 * @parameter_type int
 * @parameter_value 4
 * @parameter_description Anchor point of the rectangle window in y direction
 *
 * @parameter_name FILTER_TYPE
 * @parameter_type string
 * @parameter_value ERODE
 * @parameter_list ERODE:DILATE:OPENING:CLOSING:GRADIENT:TOPHAT:BLACKHAT
 * @parameter_description FILTER_TYPE
 *
 END*/
   
   
class CvfdhMorphologyEx : public BufferedNode {

  int m_imageInID;
  int m_imageOutID;
  int rect_size_x;
  int rect_size_y;
  int anchor_point_x;
  int anchor_point_y;
  string filter_type;
      
public:
  CvfdhMorphologyEx(string nodeName, ParameterSet params)
    : BufferedNode(nodeName, params)
  {
    m_imageInID = addInput("IMAGEIN");
    m_imageOutID = addOutput("IMAGEOUT");
    rect_size_x = dereference_cast<int>(parameters.get("RECT_SIZE_X"));
    rect_size_y = dereference_cast<int>(parameters.get("RECT_SIZE_Y"));
    anchor_point_x = dereference_cast<int>(parameters.get("ANCHOR_POINT_X"));
    anchor_point_y = dereference_cast<int>(parameters.get("ANCHOR_POINT_Y"));
    filter_type = object_cast<string>(parameters.get("FILTER_TYPE")); 
    inOrder = true;
  }      
      
  void calculate(int output_id, int count, Buffer &out)
  {
    ObjectRef in = getInput(m_imageInID, count);

    // -------------------------------------------------------------

    if (typeid(*in) == typeid(CvfdImage)) {
      
      RCPtr<CvfdImage> imagePtr = getInput(m_imageInID,count);   
      if(imagePtr->getImage() == NULL){
	(*(outputs[m_imageOutID].buffer))[count] = ObjectRef(imagePtr);
	return;
      }

      IplImage* src_img;
      src_img = cvCloneImage(imagePtr->getImage());
    
      IplConvKernel *element;
      IplImage* dst_img;
      IplImage* tmp_img;
    
      dst_img = cvCloneImage (src_img);
      tmp_img = cvCloneImage (src_img);

      element = cvCreateStructuringElementEx (rect_size_x, rect_size_y, anchor_point_x, anchor_point_y, CV_SHAPE_RECT, NULL);
    
      if(filter_type == "DILATE")
	cvDilate (src_img, dst_img, element, 1);
      if(filter_type == "ERODE")
	cvErode (src_img, dst_img, element, 1);
      if(filter_type == "OPENING")
	cvMorphologyEx (src_img, dst_img, tmp_img, element, CV_MOP_OPEN, 1);
      if(filter_type == "CLOSING")
	cvMorphologyEx (src_img, dst_img, tmp_img, element, CV_MOP_CLOSE, 1);
      if(filter_type == "GRADIENT")
	cvMorphologyEx (src_img, dst_img, tmp_img, element, CV_MOP_GRADIENT, 1);
      if(filter_type == "TOPHAT")
	cvMorphologyEx (src_img, dst_img, tmp_img, element, CV_MOP_TOPHAT, 1);
      if(filter_type == "BLACKHAT")
	cvMorphologyEx (src_img, dst_img, tmp_img, element, CV_MOP_BLACKHAT, 1);
    
      IplImage* img_out;
      img_out = cvCloneImage(dst_img);

      CvfdImage* image= new CvfdImage(img_out);  
         
      (*(outputs[m_imageOutID].buffer))[count] = ObjectRef(image); 

      cvReleaseImage (&src_img);
      cvReleaseImage (&dst_img);
      cvReleaseImage (&tmp_img);
      cvReleaseImage (&img_out);

      // -------------------------------------------------------------

    }else if (typeid(*in) == typeid(Map<int, ObjectRef>)) {

      RCPtr<Map<int, ObjectRef> > imageMapPtr = getInput(m_imageInID, count);
      RCPtr<Map<int, ObjectRef> > outputMapPtr(new Map<int, ObjectRef>);
      
      Map<int, ObjectRef>::iterator it = imageMapPtr->begin();
      for (; it != imageMapPtr->end(); ++it) {
        CvfdImage& imagePtr = object_cast<CvfdImage>(it->second);
	IplImage* src_img;
	src_img = cvCloneImage(imagePtr.getImage());

	IplConvKernel *element;
	IplImage* dst_img;
	IplImage* tmp_img;
	
	dst_img = cvCloneImage (src_img);
	tmp_img = cvCloneImage (src_img);
	
	element = cvCreateStructuringElementEx (rect_size_x, rect_size_y, anchor_point_x, anchor_point_y, CV_SHAPE_RECT, NULL);
	
	if(filter_type == "DILATE")
	  cvDilate (src_img, dst_img, element, 1);
	if(filter_type == "ERODE")
	  cvErode (src_img, dst_img, element, 1);
	if(filter_type == "OPENING")
	  cvMorphologyEx (src_img, dst_img, tmp_img, element, CV_MOP_OPEN, 1);
	if(filter_type == "CLOSING")
	  cvMorphologyEx (src_img, dst_img, tmp_img, element, CV_MOP_CLOSE, 1);
	if(filter_type == "GRADIENT")
	  cvMorphologyEx (src_img, dst_img, tmp_img, element, CV_MOP_GRADIENT, 1);
	if(filter_type == "TOPHAT")
	  cvMorphologyEx (src_img, dst_img, tmp_img, element, CV_MOP_TOPHAT, 1);
	if(filter_type == "BLACKHAT")
	  cvMorphologyEx (src_img, dst_img, tmp_img, element, CV_MOP_BLACKHAT, 1);
	
	IplImage* img_out;
	img_out = cvCloneImage(dst_img);
	
	CvfdImage* image= new CvfdImage(img_out);  

	(*outputMapPtr)[it->first] = image;

	cvReleaseImage (&src_img);
	cvReleaseImage (&dst_img);
	cvReleaseImage (&tmp_img);
	cvReleaseImage (&img_out);

      }
      (*(outputs[m_imageOutID].buffer))[count] = outputMapPtr; 
    }




  }
      
  NO_ORDER_NODE_SPEEDUP(CvfdhMorphologyEx)
};
